BEGIN {
	of1 = ARGV[2]
	of2 = ARGV[3]
	ARGV[2] = ""
	ARGV[3] = ""
	
	RS = "[ \t\n]"
	dict["+"] = "o1AOP_ADD"
	dict["-"] = "o1AOP_SUB"
	dict["/"] = "o1AOP_DIV"
	dict["%"] = "o1AOP_REM"
	dict["*"] = "o1AOP_MUL"
	dict["<<"] = "o1AOP_SHL"
	dict[">>"] = "o1AOP_SHR"

	dict["and"] = "o1AOP_AND"
	dict["&&"] = "o1AOP_LAND"
	dict["or"] = "o1AOP_OR"
	dict["||"] = "o1AOP_LOR"
	dict["xor"] = "o1AOP_XOR"
	dict["invert"] = "o2AOP_PUSHM1, AOP_XOR"
	dict["2*"] = "o2AOP_PUSH1, AOP_SHL"
	dict["2/"] = "o2AOP_PUSH1, AOP_SHR"
	
	dict["="] = "o1AOP_EQ"
	dict["!="] = "o1AOP_NE"
	dict["<"] = "o1AOP_LT"
	dict[">"] = "o1AOP_GT"
	dict["<="] = "o1AOP_LE"
	dict[">="] = "o1AOP_GE"

	dict["alloc"] = "o1AOP_ALLOC"
	dict["free"] = "o1AOP_FREE"
	dict["realloc"] = "o1AOP_REALLOC"
	dict["size"] = "o1AOP_SIZE"
	dict["!!"] = "o1AOP_STORE"
	dict["@@"] = "o1AOP_LOAD"
	dict["c!!"] = "o1AOP_STOREB"
	dict["c@@"] = "o1AOP_LOADB"
	dict["concat"] = "o1AOP_CONCAT"
	dict["duplicate"] = "o1AOP_DUPLICATE"
	dict["copyblock"] = "o1AOP_COPYBLOCK"
	dict["strcmp"] = "o1AOP_STRCMP"

	dict["getglobals"] = "o1AOP_GETGSIZE"
	dict["setglobals"] = "o1AOP_SETGSIZE"
	dict["!"] = "o1AOP_STOREG"
	dict["@"] = "o1AOP_LOADG"
	
	dict["drop"] = "o1AOP_DROP"
	dict["nip"] = "o1AOP_NIP"
	dict["dup"] = "o1AOP_DUP"
	dict["over"] = "o1AOP_OVER"
	dict["tuck"] = "o1AOP_TUCK"
	dict["swap"] = "o1AOP_SWAP"
	dict["rot"] = "o1AOP_ROT"
	dict["roll"] = "o1AOP_ROLL"
	dict["2drop"] = "o1AOP_2DROP"
	dict["2nip"] = "o1AOP_2NIP"
	dict["2dup"] = "o1AOP_2DUP"
	dict["2over"] = "o1AOP_2OVER"
	dict["2tuck"] = "o1AOP_2TUCK"
	dict["2swap"] = "o1AOP_2SWAP"
	dict["2rot"] = "o1AOP_2ROT"
	
	dict["pick"] = "o1AOP_PICK"
	dict["?dup"] = "o1AOP_QDUP"

	dict[">R"] = "o1AOP_PUSHC"
	dict["R>"] = "o1AOP_POPC"
	
	dict["stop"] = "o1AOP_STOP"
	dict["nextevent"] = "o1AOP_NEXTEVENT"
	dict["ijump"] = "o1AOP_IJUMP"

	dict["module"] = "o1AOP_MODULE"
	dict["mexec"] = "o1AOP_MEXEC"
	
	dict["debugs"] = "o1AOP_DEBUGS"
	dict["debugi"] = "o1AOP_DEBUGI"
	dict["debugf"] = "o1AOP_DEBUGF"
	
	pc = 0
	label = 0
	sp = 0
	lp = 0
	
	bp = -1
}

function pushconst(val) {
	if (val == 0)
	{
		print "AOP_PUSH0,"  > of1
		pc += 1
	}
	else if (val == 1)
	{
		print "AOP_PUSH1," > of1
		pc += 1
	}
	else if (val == -1)
	{
		print "AOP_PUSHM1," > of1
		pc += 1
	}
	else if ((val >= -128) && (val <= 127))
	{
		print "AOP_PUSH8, "val" ," > of1
		pc += 2
	}
	else if ((val >= -32768) && (val <= 32767))
	{
		print "AOP_PUSH16, AOP_I16(" val ")," > of1
		pc += 3
	}
	else
	{
		print "AOP_PUSH32, AOP_I32(" val ")," > of1
		pc += 5
	}
}

function munge(s) {
	gsub(/!/, "_pling_", s)
	gsub(/@/, "_at_", s)
	gsub(/-/, "_dash_", s)
	gsub(/\./, "_dot_", s)
	return s;
}

/^"/ {
	s = $0
	gsub(/^"/, "", s)
	if (s ~ /"$/)
		gsub(/"$/, "", s)
	else
	{
		while (getline t)
		{
			s = s " " t
			if (t ~ /"$/)
			{
				gsub(/"$/, "", s)
				break
			}
		}
	}
	print "AOP_ALLOCCONST, AOP_I16(" length(s) ")," > of1
	for(i=1; i<=length(s); i++)
	{
		c = substr(s, i, 1)
		if (c == "\\")
			c = "\\\\"
		if (c == "'")
			c = "\\'"
		printf "%s", "'" c "', " > of1
	}
	print "" > of1
	pc += 3 + length(s)
	next
}

{
	def = dict[$0]
	if (def == "")
	{
		if (($0 ~ /^-*[0-9]+$/) || ($0 ~ /^-*0x-*[0-9a-fA-F]+$/))
			pushconst($0)
		else if ($0 == "include")
		{
			getline name
			ARGV[ARGIND] = name
		}
		else if ($0 == "try")
		{
			stack[sp] = "_" label
			print "/* try */" > of1
			print "AOP_PUSH16, AOP_I16(_" label "), AOP_MARKEX," > of1
			pc += 4
			sp += 1
			lp += 3
			label += 1
		}
		else if ($0 == "catch")
		{
			print "AOP_JUMP, AOP_I16(" stack[sp-1] "_exit)," > of1
			pc += 3
			print "/* catch */" > of1
			print "/* " stack[sp-1] " */" > of1
			dict[" " stack[sp-1]] = "a" pc
		}
		else if ($0 == "endtry")
		{
			print "/* endtry */" > of1
			print "/* " stack[sp-1] "_exit */" > of1
			dict[" " stack[sp-1] "_exit"] = "a" pc
			print "AOP_RESTOREEX," > of1
			pc += 1
			lp -= 3
			sp -= 1
		}
		else if ($0 == "local")
		{
			getline name
			dict[name] = "l" lp
			print "/* local " name " @ " lp " */" > of1
			print "AOP_PUSH0, AOP_PUSHC," > of1
			pc += 2
			lp += 1
		}
		else if ($0 == "ilocal")
		{
			getline name
			dict[name] = "l" lp
			print "/* ilocal " name " @ " lp " */" > of1
			print "AOP_PUSHC," > of1
			pc += 1
			lp += 1
		}
		else if ($0 == "TO")
		{
			getline name
			def = dict[name]
			if (def !~ /^l/)
			{
				print name " is not a defined local"
				exit 1
			}
			gsub(/^l/, "", def)
			print "/* local store " $0 " @ " def ", lp=" lp " */" > of1
			pushconst(lp - def)
			print "AOP_STOREC," > of1
			pc += 1
		}
		else if ($0 == "[ADDRESS]")
		{
			getline word
			print "AOP_PUSH16, AOP_I16(" munge(word) ")," > of1
			pc += 3
		}
		else if ($0 == "exit")
		{
			if (lp)
			{
				pushconst(lp)
				print "AOP_RETRACTC," > of1
				pc += 1
			}
			print "AOP_RETURN," > of1
			pc += 1
		}
		else if ($0 == ";")
		{
			if (lp)
			{
				pushconst(lp)
				print "AOP_RETRACTC," > of1
				pc += 1
			}
			print "AOP_RETURN," > of1
			pc += 1
			lp = 0
		}
		else if ($0 == "(")
		{
			do {
				getline n
			} while (n != ")")
		}
		else if ($0 == "if")
		{
			stack[sp] = "_" label
			print "AOP_NOT, AOP_CJUMP, AOP_I16(_" label ")," > of1
			pc += 4
			sp += 1
			label += 1
		}
		else if ($0 == "else")
		{
			print "AOP_JUMP, AOP_I16(_" label ")," > of1
			pc += 3
			print "/* else " stack[sp-1] " */" > of1
			dict[" " stack[sp-1]] = "a" pc
			stack[sp-1] = "_" label
			label += 1
		}
		else if ($0 == "then")
		{
			sp -= 1
			print "/* " stack[sp] " */" > of1
			dict[" " stack[sp]] = "a" pc
		}
		else if ($0 == "begin")
		{
			stack[sp] = bp
			sp += 1
			bp = sp
			stack[sp] = "_" label
			print "/* " stack[sp] " */" > of1
			dict[" " stack[sp]] = "a" pc
			sp += 1
			label += 1
		}
		else if ($0 == "again")
		{
			sp -= 1
			print "AOP_JUMP, AOP_I16(" stack[sp] ")," > of1
			sp -= 1
			bp = stack[sp]
			pc += 3
			print "/* " stack[sp+1] "_exit */" > of1
			dict[" " stack[sp+1] "_exit"] = "a" pc
		}
		else if ($0 == "leave")
		{
			print "AOP_JUMP, AOP_I16(" stack[bp] "_exit)," > of1
			pc += 3
		}
		else if ($0 == "constant")
		{
			getline name
			getline val
			dict[name] = "c" val
			print "/* constant " name " = " val " */" > of1
		}
		else if ($0 == "jump")
		{
			getline dest
			print "AOP_JUMP, AOP_I16(" dest ")," > of1
			pc += 3
		}
		else if ($0 == ":")
		{
			getline name
			dict[name] = "a" pc
			print "/* word " name " */" > of1
		}
		else if ($0 == "")
		{
		}
		else
		{
			name = munge($0)
			print "AOP_CALL, AOP_I16(" munge(name) ")," > of1
			pc += 3
		}
	}
	else if (def ~ /^l/)
	{
		gsub(/^l/, "", def)
		print "/* local fetch " $0 " @ " def ", lp=" lp " */" > of1
		pushconst(lp - def)
		print "AOP_LOADC," > of1
		pc += 1
	}
	else if (def ~ /^o/)
	{
		gsub(/^o/, "", def)
		pc += substr(def, 1, 1)
		gsub(/^[0-9]/, "", def)
		print def "," > of1
	}
	else if (def ~ /^a/)
	{
		print "AOP_CALL, AOP_I16(" munge($0) ")," > of1
		pc += 3
	}
	else if (def ~ /^c/)
	{
		gsub(/^c/, "", def)
		pushconst(def)
	}
}

END {
	for(i in dict)
	{
		j = dict[i]
		if (j ~ /^a/)
		{
			gsub(/^a/, "", j)
			print munge(i) " = " j "," > of2
		}
	}
}

